""" add a keycloak authentication class specific to Django Rest Framework """
from typing import Tuple, Dict, List
import logging
import re

from django.core.exceptions import ObjectDoesNotExist, FieldError
from django.contrib.auth.models import AnonymousUser, update_last_login, Group
from django.contrib.auth import get_user_model
from rest_framework import authentication
from rest_framework.exceptions import AuthenticationFailed
from .keycloak import (
    OIDCConfigException,
    get_keycloak_openid,
    get_resource_roles,
    add_roles_prefix,
)
from .settings import api_settings
from .utils import get_token_issuer
from . import __title__


log = logging.getLogger(__title__)
User = get_user_model()


class KeycloakAuthentication(authentication.TokenAuthentication):
    keyword = api_settings.KEYCLOAK_AUTH_HEADER_PREFIX

    keycloak_openid = None

    def authenticate(self, request):
        credentials = super().authenticate(request)
        if credentials:
            user, decoded_token = credentials

            # Append realm_name
            decoded_token.update({"realm_name": self.keycloak_openid.realm_name})

            # Expose Keycloak roles
            request.roles = self._get_roles(user, decoded_token)

            # Expose keycloak_openid
            request.keycloak_openid = self.keycloak_openid

            # Create local application groups?
            if api_settings.KEYCLOAK_MANAGE_LOCAL_GROUPS is True:
                groups = self._get_or_create_groups(request.roles)
                user.groups.set(groups)

            # Set user is_staff based on Keycloak role mapping
            self._user_toggle_is_staff(request, user)

        return credentials

    def authenticate_credentials(self, key: str) -> Tuple[AnonymousUser, Dict]:
        """Attempt to verify JWT from Authorization header with Keycloak"""
        log.debug("KeycloakAuthentication.authenticate_credentials")
        try:
            # Create a default KeycloakOpenID configuration if not already available
            if self.keycloak_openid is None:
                self.keycloak_openid = get_keycloak_openid()

            user = None
            # Checks token is active
            decoded_token = self._get_decoded_token(key)
            self._verify_token_active(decoded_token)
            if api_settings.KEYCLOAK_MANAGE_LOCAL_USER is not True:
                log.debug(
                    "KeycloakAuthentication.authenticate_credentials: "
                    f"{decoded_token}"
                )
                user = AnonymousUser()
            else:
                user = self._handle_local_user(decoded_token)

            log.debug(
                "KeycloakAuthentication.authenticate_credentials:\n"
                "################# decoded_token ###############\n"
                f"{user} | {decoded_token}\n"
                "################ /decoded_token ###############"
            )

            return (user, decoded_token)
        except Exception as e:
            log.error(
                "KeycloakAuthentication.authenticate_credentials | " f"Exception: {e}"
            )
            raise AuthenticationFailed() from e

    def _get_decoded_token(self, token: str) -> dict:
        return self.keycloak_openid.introspect(token)

    def _verify_token_active(self, decoded_token: dict) -> None:
        """raises if not active"""
        is_active = decoded_token.get("active", False)
        if not is_active:
            raise AuthenticationFailed("Invalid or expired token")

    def _add_realm_prefix(self, value: str) -> str:
        """Add 'REALM_NAME:' prefix to a string value.
        This only works if using KEYCLOAK_MULTI_OIDC_JSON

        Checks to ensure the prefix is not already added.
        :param value: The string to prefix
        :returns: Prefixed string
        """
        if not api_settings.KEYCLOAK_MULTI_OIDC_JSON:
            return value

        if not self.keycloak_openid.realm_name:
            log.warning(f"Cannot add realm prefix. Realm missing!")
            return value

        prefix = str(self.keycloak_openid.realm_name) + ":"
        if re.search("^" + prefix, value):
            log.debug(f"Value '{str(value)}' already has realm prefix")
            return value

        return prefix + str(value)

    def _map_keycloak_to_django_fields(self, decoded_token: dict) -> dict:
        """Map Keycloak access_token fields to Django User attributes"""
        django_fields = {}
        kc_username_field = api_settings.KEYCLOAK_FIELD_AS_DJANGO_USERNAME

        if kc_username_field and isinstance(kc_username_field, str):
            django_fields["username"] = self._add_realm_prefix(
                decoded_token.get(kc_username_field, "")
            )

        # ecocommons_user.CustomUser field
        django_fields["realm"] = str(self.keycloak_openid.realm_name)

        django_fields["email"] = decoded_token.get("email", "")

        # django stores first_name and last_name as empty strings
        # by default, not None
        django_fields["first_name"] = decoded_token.get("given_name", "")
        django_fields["last_name"] = decoded_token.get("family_name", "")

        return django_fields

    def _update_user(self, user: User, django_fields: dict) -> User:
        """if user exists, keep data updated as necessary"""
        save_model = False

        for key, value in django_fields.items():
            try:
                if getattr(user, key) != value:
                    setattr(user, key, value)
                    save_model = True
            except Exception:
                log.warning(
                    "KeycloakAuthentication."
                    "_update_user | "
                    f"setattr: {key} field does not exist"
                )
        if save_model:
            user.save()
        return user

    def _handle_local_user(self, decoded_token: dict) -> User:
        """ used to update/create local users from keycloak data """
        django_uuid_field = api_settings.KEYCLOAK_DJANGO_USER_UUID_FIELD

        sub = decoded_token["sub"]
        django_fields = self._map_keycloak_to_django_fields(decoded_token)

        user = None
        try:
            user = User.objects.get(**{django_uuid_field: sub})
            user = self._update_user(user, django_fields)

        except ObjectDoesNotExist:
            log.warning(
                "KeycloakAuthentication._handle_local_user | "
                f"ObjectDoesNotExist: {sub} does not exist - creating"
            )
        except User.DoesNotExist:
            print(f"ERROR: User with {django_uuid_field}={sub} not found!")
        except FieldError:
            print(f"ERROR: Invalid field {django_uuid_field} in User model.")
        except Exception as e:
            print(f"ERROR: Unexpected error - {e}")

        if user is None:
            # Add uuid field and create
            django_fields.update(**{django_uuid_field: sub})
            user = User.objects.create_user(**django_fields)

        update_last_login(sender=None, user=user)
        return user

    def _get_roles(self, user: User, decoded_token: dict) -> List[str]:
        """try to add roles from authenticated keycloak user"""
        roles = []
        try:
            roles += get_resource_roles(decoded_token, self.keycloak_openid.client_id)
            roles.append(str(user.pk))
        except Exception as e:
            log.warning("KeycloakAuthentication._get_roles | " f"Exception: {e}")

        log.debug(f"KeycloakAuthentication._get_roles: {roles}")
        return roles

    def _get_or_create_groups(self, roles: List[str]) -> List[Group]:
        groups = []
        for role in roles:
            group, created = Group.objects.get_or_create(name=role)
            if created:
                log.debug(
                    "KeycloakAuthentication._get_or_create_groups | created: "
                    f"{group.name}"
                )
            else:
                log.debug(
                    "KeycloakAuthentication._get_or_create_groups | exists: "
                    f"{group.name}"
                )
            groups.append(group)
        return groups

    def _user_toggle_is_staff(self, request, user: User) -> None:
        """
        toggle user.is_staff if a role mapping has been declared in settings
        """
        try:
            # catch None or django.contrib.auth.models.AnonymousUser
            valid_user = bool(
                user
                and isinstance(user, User)
                and hasattr(user, "is_staff")
                and getattr(user, "is_superuser", False) is False
            )
            log.debug(
                f"KeycloakAuthentication._user_toggle_is_staff | {user} | "
                f"valid_user: {valid_user}"
            )
            if (
                valid_user
                and api_settings.KEYCLOAK_ROLES_TO_DJANGO_IS_STAFF
                and type(api_settings.KEYCLOAK_ROLES_TO_DJANGO_IS_STAFF)
                in [list, tuple, set]
            ):
                is_staff_roles = set(
                    add_roles_prefix(api_settings.KEYCLOAK_ROLES_TO_DJANGO_IS_STAFF)
                )
                log.debug(
                    f"KeycloakAuthentication._user_toggle_is_staff | {user} | "
                    f"is_staff_roles: {is_staff_roles}"
                )
                user_roles = set(request.roles)
                log.debug(
                    f"KeycloakAuthentication._user_toggle_is_staff | {user} | "
                    f"user_roles: {user_roles}"
                )
                is_staff = bool(is_staff_roles.intersection(user_roles))
                log.debug(
                    f"KeycloakAuthentication._user_toggle_is_staff | {user} | "
                    f"is_staff: {is_staff}"
                )
                # don't write unnecessarily, check different first
                if is_staff != user.is_staff:
                    user.is_staff = is_staff
                    user.save()

        except Exception as e:
            log.warning(
                "KeycloakAuthentication._user_toggle_is_staff | " f"Exception: {e}"
            )


class KeycloakMultiAuthentication(KeycloakAuthentication):
    """Keycloak authentication for multiple realms. Determined by KEYCLOAK_MULTI_OIDC_JSON"""

    def authenticate_credentials(self, key: str) -> Tuple[AnonymousUser, Dict]:
        """
        Decode the unverified token to get the issuer and validate it against ALLOWED_HOSTS.
        This determines the auth realm
        """
        try:
            self.keycloak_openid = get_keycloak_openid(host=get_token_issuer(key))
            return super().authenticate_credentials(key)

        except OIDCConfigException as e:
            log.warning(
                "KeycloakMultiAuthentication.authenticate | "
                f"OIDCConfigException: {e})"
            )
            raise AuthenticationFailed() from e

        except AuthenticationFailed as e:
            log.info(
                "KeycloakMultiAuthentication.authenticate | "
                f"AuthenticationFailed: {e})"
            )
            raise e

        except Exception as e:
            log.error("KeycloakMultiAuthentication.authenticate | " f"Exception: {e})")
            raise AuthenticationFailed() from e


class KeycloakSessionAuthentication(KeycloakAuthentication):

    def authenticate_credentials(self, key: str) -> Tuple[AnonymousUser, Dict]:
        """
        Decode the unverified token to get the issuer and validate it against ALLOWED_HOSTS.
        This determines the auth realm
        """
        try:
            self.keycloak_openid = get_keycloak_openid(host=get_token_issuer(key))
            return super().authenticate_credentials(key)

        except OIDCConfigException as e:
            log.warning(
                "KeycloakSessionAuthentication.authenticate | "
                f"OIDCConfigException: {e})"
            )
            raise AuthenticationFailed() from e

        except AuthenticationFailed as e:
            log.info(
                "KeycloakSessionAuthentication.authenticate | "
                f"AuthenticationFailed: {e})"
            )
            raise e

        except Exception as e:
            log.error(
                "KeycloakSessionAuthentication.authenticate | " f"Exception: {e})"
            )
            raise AuthenticationFailed() from e



""" module for app specific keycloak connection """
from typing import Dict, List
import traceback
import logging

from keycloak import KeycloakOpenID

from .settings import api_settings
from . import __title__


log = logging.getLogger(__title__)


class OIDCConfigException(Exception):
    pass


def get_request_oidc_config(host: str) -> dict:
    """ Determine client config from request host.
        :param host: Hostname

        KEYCLOAK_MULTI_OIDC_JSON: {
            "server": {
                "auth-server-url":  "https://server/auth/",
                "realm":            "realm_name",
                "resource":         "client_id",
                "credentials": {
                    "secret": "client_secret"
                }
            }
        }
    """

    def get_host_oidc(hostname: str, oidc_config_dict: dict) -> dict:
        if isinstance(oidc_config_dict, dict):
            log.debug(f"get_host_oidc: looking up 'oidc_config_dict' for hostname ({hostname})")
        for key, config in oidc_config_dict.items():
            if key in str(hostname) or hostname == key:
                log.debug(f"get_host_oidc: Found OIDC adapter for '{hostname}'")
                return config
        return None

    if not isinstance(host, str):
        raise OIDCConfigException(f"Cannot determine OIDC config. Missing 'host'")

    if not isinstance(api_settings.KEYCLOAK_MULTI_OIDC_JSON, dict):
        raise OIDCConfigException(f"OIDC config KEYCLOAK_MULTI_OIDC_JSON not available")

    oidc_config = get_host_oidc(
        host,
        api_settings.KEYCLOAK_MULTI_OIDC_JSON
    )

    if oidc_config is None:
        raise OIDCConfigException(f"Could not determine OIDC config for "
                                  f"'{str(host)}'")

    return oidc_config


def get_keycloak_openid(host: str = None) -> KeycloakOpenID:
    """ 
        Create a KeycloakOpenID instance from application credentials.
        :param host: If request host is provided will attempt to set application credentials 
                        based on a predefined host list.
                            KEYCLOAK_MULTI_OIDC_JSON

                        otherwise it will fallback to the default configured application credentials
                            KEYCLOAK_SERVER_URL, 
                            KEYCLOAK_REALM, 
                            KEYCLOAK_CLIENT_ID, 
                            KEYCLOAK_CLIENT_SECRET_KEY

        :returns: KeycloakOpenID
    """
    try:
        if isinstance(host, str):
            try: 
                oidc_config = get_request_oidc_config(host)
                if oidc_config:
                    log.debug(
                        'get_keycloak_openid: '
                        f'OIDC realm={oidc_config["realm"]}'
                    )
                    return KeycloakOpenID(
                        server_url=oidc_config["auth-server-url"],
                        realm_name=oidc_config["realm"],
                        client_id=oidc_config["resource"],
                        client_secret_key=oidc_config["credentials"]["secret"]
                    )

            except OIDCConfigException as e:
                if not api_settings.KEYCLOAK_SERVER_URL:
                    raise e

        return KeycloakOpenID(
            server_url=api_settings.KEYCLOAK_SERVER_URL,
            realm_name=api_settings.KEYCLOAK_REALM,
            client_id=api_settings.KEYCLOAK_CLIENT_ID,
            client_secret_key=api_settings.KEYCLOAK_CLIENT_SECRET_KEY
        )

    except KeyError as e:
        raise KeyError(
            f'invalid settings: {e}'
        ) from e


def get_resource_roles(decoded_token: Dict, client_id=None) -> List[str]:
    """ Get roles from access token """
    resource_access_roles = []
    try:
        if client_id is None:
            client_id = api_settings.KEYCLOAK_CLIENT_ID

        log.debug(f'{__name__} - get_resource_roles - client_id: {client_id}')

        resource_access_roles = (
            decoded_token
            .get('resource_access', {})
            .get(client_id, {})
            .get('roles', [])
        )
        roles = add_roles_prefix(resource_access_roles)
        log.debug(f'{__name__} - get_resource_roles - roles: {roles}')

        return roles

    except Exception as e:
        log.warning(f'{__name__} - get_resource_roles - Exception: ({str(type(e).__name__ )}) {e}\n'
                    f'{traceback.format_exc()}')
        return []


def add_roles_prefix(roles: List[str]) -> List[str]:
    """ add role prefix configured by KEYCLOAK_ROLE_SET_PREFIX to a list of roles """
    log.debug(f'{__name__} - get_resource_roles - roles: {roles}')
    prefixed_roles = [prefix_role(x) for x in roles]
    log.debug(
        f'{__name__} - get_resource_roles - prefixed_roles: {prefixed_roles}'
    )
    return prefixed_roles


def prefix_role(role: str) -> str:
    """ add prefix to role string """
    role_prefix = (
        api_settings.KEYCLOAK_ROLE_SET_PREFIX
        if api_settings.KEYCLOAK_ROLE_SET_PREFIX
        and isinstance(api_settings.KEYCLOAK_ROLE_SET_PREFIX, str)
        else ''
    )
    return f'{role_prefix}{role}'



import logging
from typing import List

from rest_framework import permissions
from django.contrib.auth.models import AbstractUser

from . import __title__
from .keycloak import prefix_role

log = logging.getLogger(__title__)

ROLE_GUEST = prefix_role('guest')
ROLE_USER = prefix_role('user')
ROLE_SERVICE = prefix_role('service')
ROLE_ADMIN = prefix_role('admin')


def _has_required_group(request, required_groups: List[str]) -> bool:
    user_groups = [x.name for x in request.user.groups.all()]
    log.info(
        '_has_required_group | required_groups: '
        f'{required_groups}'
    )
    log.info(
        '_has_required_group | user_groups: '
        f'{user_groups}'
    )
    return bool(
        set(required_groups)
        .intersection(
            set([x.name for x in request.user.groups.all()])
        )
    )


class BaseGroupBasedPermission(permissions.BasePermission):
    required_groups = []

    def has_permission(self, request, view):
        log.info(
            'BaseGroupBasedPermission.has_permission: '
            f'{self.required_groups}'
        )
        return bool(
            request.user and _has_required_group(request, self.required_groups)
        )

    def has_object_permission(self, request, view, obj):
        log.info('BaseGroupBasedPermission.has_object_permission')
        return self.has_permission(request, view)


class HasAdminGroup(BaseGroupBasedPermission):
    required_groups = [ROLE_ADMIN]


class HasUserGroup(BaseGroupBasedPermission):
    required_groups = [ROLE_USER]


class HasServiceGroup(BaseGroupBasedPermission):
    required_groups = [ROLE_SERVICE]


class HasOwnerGroup(permissions.BasePermission):
    """ validate auth user is obj.owner """

    def has_object_permission(self, request, view, obj):
        required_groups = [str(obj.owner.pk)]
        log.info(
            'HasOwnerGroup.has_object_permission: '
            f'{required_groups}'
        )
        return bool(
            request.user and _has_required_group(request, required_groups)
        )


def _has_required_role(request, required_roles: List[str]) -> bool:
    log.info(
        '_has_required_role | required_roles: '
        f'{required_roles}'
    )
    roles = getattr(request, 'roles', [])
    log.info(
        '_has_required_role | request.roles: '
        f'{roles}'
    )
    return bool(set(required_roles).intersection(set(roles)))


class BaseRoleBasedPermission(permissions.BasePermission):
    required_roles = []

    def has_permission(self, request, view):
        log.info(
            'BaseRoleBasedPermission.has_permission: '
            f'{self.required_roles}'
        )
        return bool(
            request.user and _has_required_role(request, self.required_roles)
        )

    def has_object_permission(self, request, view, obj):
        log.info('BaseRoleBasedPermission.has_object_permission')
        return self.has_permission(request, view)


class HasAdminRole(BaseRoleBasedPermission):
    required_roles = [ROLE_ADMIN]


class HasUserRole(BaseRoleBasedPermission):
    required_roles = [ROLE_USER]


class HasServiceRole(BaseRoleBasedPermission):
    required_roles = [ROLE_SERVICE]


class HasOwnerRole(permissions.BasePermission):
    """ validate auth user is obj.owner """

    def has_object_permission(self, request, view, obj):
        ownerid = obj.owner.pk if isinstance(obj.owner, AbstractUser) else obj.owner
        required_roles = [str(ownerid)]
        log.info(
            'HasOwnerRole.has_object_permission: '
            f'{required_roles}'
        )
        return bool(
            request.user and _has_required_role(request, required_roles)
        )



""" Settings config for the drf_keycloak_auth application """
import os

from django.conf import settings
from rest_framework.settings import APISettings

import json

USER_SETTINGS = getattr(settings, 'DRF_KEYCLOAK_AUTH', None)

# should be comma separated string
KEYCLOAK_ROLES_TO_DJANGO_IS_STAFF = \
    os.getenv('KEYCLOAK_ROLES_TO_DJANGO_IS_STAFF')

DEFAULTS = {
    'ALLOWED_HOSTS': os.getenv('DJANGO_ALLOWED_HOSTS', '').split(' '),

    'KEYCLOAK_MULTI_OIDC_JSON': (
        json.loads(os.getenv('KEYCLOAK_MULTI_OIDC_JSON'))
        if os.getenv('KEYCLOAK_MULTI_OIDC_JSON')
        else None
    ),
    
    'KEYCLOAK_SERVER_URL': os.getenv('KEYCLOAK_SERVER_URL'),

    'KEYCLOAK_REALM': os.getenv('KEYCLOAK_REALM'),

    'KEYCLOAK_CLIENT_ID': os.getenv('KEYCLOAK_CLIENT_ID'),

    'KEYCLOAK_CLIENT_SECRET_KEY': os.getenv('KEYCLOAK_CLIENT_SECRET_KEY'),

    'KEYCLOAK_AUTH_HEADER_PREFIX':
        os.getenv('KEYCLOAK_AUTH_HEADER_PREFIX', 'Bearer'),

    'KEYCLOAK_ROLE_SET_PREFIX':
        os.getenv('KEYCLOAK_ROLE_SET_PREFIX', 'role:'),

    'KEYCLOAK_MANAGE_LOCAL_USER':
        os.getenv('KEYCLOAK_MANAGE_LOCAL_USER', 'true').lower() in ('true', '1'),

    'KEYCLOAK_MANAGE_LOCAL_GROUPS':
        os.getenv('KEYCLOAK_MANAGE_LOCAL_GROUPS', 'false').lower() in ('true', '1'),

    'KEYCLOAK_DJANGO_USER_UUID_FIELD':
        os.getenv('KEYCLOAK_DJANGO_USER_UUID_FIELD', 'pk'),

    'KEYCLOAK_FIELD_AS_DJANGO_USERNAME':
        os.getenv('KEYCLOAK_FIELD_AS_DJANGO_USERNAME', 'preferred_username'),

    'KEYCLOAK_ROLES_TO_DJANGO_IS_STAFF': (
        [x.strip() for x in KEYCLOAK_ROLES_TO_DJANGO_IS_STAFF.split(',')]
        if KEYCLOAK_ROLES_TO_DJANGO_IS_STAFF
        else ['admin']  # can be list, tuple or set
    )
}

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',

    'drf_keycloak_auth.middleware.LoginRequiredMiddleware',
]

# List of settings that may be in string import notation.
IMPORT_STRINGS = (
)

api_settings = APISettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS)



import os
from requests_oauthlib import OAuth2Session
from oauthlib.oauth2 import BackendApplicationClient
from keycloak import KeycloakOpenID
import logging

log = logging.getLogger('BackendRequestClient')

class BackendRequestClientException(Exception):
    pass

class BackendRequestClient:
    """ Backend OAuth2 client for use with client credentials. """

    def __init__(self, keycloak_openid: KeycloakOpenID):
        """ 
            :param keycloak_openid: KeycloakOpenID
        """
        if not keycloak_openid:
            raise BackendRequestClientException("KeycloakOpenID not provided")

        self.keycloak_openid = keycloak_openid
        self.token_endpoint = self.get_token_url(
            keycloak_openid.connection.base_url,
            keycloak_openid.realm_name
        )
        self.client = BackendApplicationClient(client_id=keycloak_openid.client_id)
        self.session = OAuth2Session(client=self.client)

    def get_token_url(self, host: str, realm_name: str) -> str:
        """ OAuth token endpoint 
            :param host: Auth provider host
            :param realm_name: Realm name
            :returns: Token URL
        """
        return os.path.join(f"{host}",
                            f"realms",
                            f"{realm_name}",
                            f"protocol/openid-connect/token")

    def fetch_token(self) -> dict:
        log.info(f"fetch_token: {self.token_endpoint}")

        return self.session.fetch_token(
            token_url=self.token_endpoint,
            client_id=self.keycloak_openid.client_id,
            client_secret=self.keycloak_openid.client_secret_key,
            scope=["openid", "profile"]
        )

    def get(self, url):
        """ GET request method """
        self.fetch_token()
        return self.session.get(url)

    def post(self, url, json):
        """ POST request method """
        self.fetch_token()
        return self.session.post(url, json=json)


import os 
import requests
from django.shortcuts import redirect
from django.http import JsonResponse
from drf_keycloak_auth.authentication import (
    KeycloakSessionAuthentication,
)
from drf_keycloak_auth.keycloak import get_keycloak_openid

import logging

log = logging.getLogger("LoginRequiredMiddleware")


class LoginRequiredMiddleware:
    """Middleware that forces authentication using DRF authentication"""

    def __init__(self, get_response):
        self.get_response = get_response
        self.authenticator = KeycloakSessionAuthentication()  # Use DRF authentication
        self.keycloak_openid = get_keycloak_openid()  # Initialize KeycloakOpenID

    def __call__(self, request):

        if "text/html" in request.headers.get("Accept", "").lower():
            code = request.GET.get("code")
            base_url = f"{request.scheme}://{request.get_host()}"

            if not code:
                login_url = self.keycloak_openid.auth_url(redirect_uri=base_url)
                return redirect(login_url)

            # If code exists, exchange it for the tokens
            try:
                # Exchange code for tokens

                token_url = os.path.join(
                    f"{self.keycloak_openid.connection.base_url}",
                    "realms",
                    f"{self.keycloak_openid.realm_name}",
                    "protocol/openid-connect/token"
                )

                response = requests.post(
                    token_url,
                    data={
                        "client_id": self.keycloak_openid.client_id,
                        "client_secret": self.keycloak_openid.client_secret_key,
                        "code": code,
                        "grant_type": "authorization_code",
                        "redirect_uri": base_url,
                    },
                    headers={
                        "Content-Type": "application/x-www-form-urlencoded",
                        "Accept": "application/json",
                    },
                )

                if response.status_code == 200:
                    tokens = response.json()
                    access_token = tokens.get("access_token")

                    # Optionally, use the access token to authenticate the user
                    user = self.authenticator.authenticate_credentials(
                        access_token
                    )
                    request.user = user  # Assign authenticated user to request object

                else:
                    raise Exception(
                        f"Failed to get tokens from Keycloak: {response.text}"
                    )

            except Exception as e:
                log.error(f"Error during authentication: {str(e)}")  # Log any errors
                return JsonResponse({"error": "Authentication failed"}, status=401)

        # Continue with the request handling
        response = self.get_response(request)
        return response



#!/usr/bin/env python3
from os import path
import setuptools

import drf_keycloak_auth as meta

this_dir = path.abspath(path.dirname(__file__))
with open(path.join(this_dir, 'README.md'), mode='r') as f:
    long_description = f.read()

setuptools.setup(
    name=meta.__title__,
    version=meta.__version__,
    author=meta.__author__,
    author_email=meta.__author_email__,
    description=meta.__description__,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=meta.__url__,
    license=meta.__license__,
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
    install_requires=[
        'djangorestframework>=3.15,<4',
        'python-keycloak>=5.1',
        'django>=4.1',
        'requests_oauthlib>=2.0.0',
        'urllib3<1.27,>=1.21.1',
        'pyjwt>=2.6',
        'pytz'
    ],
    keywords=[
        'drf',
        'django',
        'rest_framework',
        'djangorestframework',
        'authentication',
        'python3',
        'keycloak'
    ],
)
